/*
 *  remailckd.c
 *
 *  An RFC 1339 remote mailcheck daemon for maildirs (without authentication).
 * 
 *  To have remailck work on an account, it must have a ~/Maildir with
 *  mode o=rx, and ~/Maildir/{new,cur} with mode o=r.  That way the
 *  daemon can see if there is mail without being able to read it 
 *
 *
 *  Copyright (C) 2002 Tim Deegan (tjd@phlegethon.org)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 *  $Id: remailckd.c,v 1.8 2002/06/14 14:29:45 tjd Rel $
 *
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <dirent.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <utmp.h>

/* Max. request rate to handle */
#define RATE_INTERVAL  5 /* seconds */
#define RATE_COUNT    30 /* requests per RATE_INTERVAL */

/* Port to run on */
#define REMAILCK_PORT 50

/* How long is a login name on this system? */
#ifndef UT_NAMESIZE
#define UT_NAMESIZE 8
#endif

/* Solaris doesn't define this */
#ifndef INADDR_NONE
#define INADDR_NONE (in_addr_t)0xffffffff
#endif


/* What characters are acceptable in a username? */
#define ACCEPTABLE_CHARS \
 "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_"

/* Standard responses */
static const char no_mail[12]  = {0,0,0,0,0,0,0,0,0,0,0,0};
static const char new_mail[12] = {0,0,0,0,0,0,0,1,0,0,0,0};
static const char old_mail[12] = {0,0,0,0,0,0,0,0,0,0,0,1};

/* Forward definitions */
static void failure(const char *why, ...);
static int open_socket_fifty(const char *addr_string);
static void drop_privilege(const char *user);
static void serviceloop(int socket50);


/*  |
 *  |  You are here
 *  V  
 */
int main (int argc, char **argv)
{
    int socket50;

    /* WARNING: Ths code is run as root */
    
    /* Check that we are root */
    if (getuid() != 0 && geteuid() != 0) {
	failure("must be run as root.\n"
		"usage: remailckd <user> <ip>\n\n"
		"Runs an RFC 1339 daemon listening on <ip> as user <user>.\n");
    }

    /* Check the arguments */
    if (argc != 3 
	|| argv[0] == NULL || argv[1] == NULL || argv[2] == NULL
	|| argv[3] != NULL) {
	drop_privilege("nobody");

	failure("usage: remailckd <user> <ip>\n\n"
		"Runs an RFC 1339 daemon listening on <ip> as user <user>.\n");

    }
	
    /* Get the root stuff out of the way first */    
    socket50 = open_socket_fifty(argv[2]);
    drop_privilege(argv[1]);

    /* Play nice with others */
    if (setpriority(PRIO_PROCESS, 0, 20) != 0)
	failure("cannot renice myself (%s)", strerror(errno));

    /* And away we go */
    serviceloop(socket50);

    return 0;
}



/*  
 *  Helper functions that run as root 
 */

static int open_socket_fifty(const char *addr_string)
    /* Open a UDP socket on port 50, or exit with an error message.
     * WARNING: This code is run as root. */
{
    int s;
    struct in_addr addr;
    struct sockaddr_in sockaddr;

    /* Parse the requested IP address */
    if (addr_string == NULL 
	|| addr_string[0] == '\0' 
	|| strlen(addr_string) > 15
	|| (addr.s_addr = inet_addr(addr_string)) == INADDR_NONE)
    {
	drop_privilege("nobody");
	failure("cannot parse \"%s\" as an IP address.", addr_string);
    }
    
    /* Open a UDP socket */
    s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (s < 0) {
	drop_privilege("nobody");
	failure("cannot open a UDP socket (%s).", strerror(errno));
    }
    
    /* Bind to address and port */
    memset(&sockaddr, 0, sizeof(sockaddr));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = htons(REMAILCK_PORT);
    sockaddr.sin_addr = addr;
    if (bind(s, (struct sockaddr *)&sockaddr, sizeof(sockaddr)) != 0)
    {
	drop_privilege("nobody");
	failure("cannot bind to address %s:%i (%s).", 
		inet_ntoa(addr), REMAILCK_PORT, strerror(errno));
    }

    /* Success */
    return (s);
}




static void drop_privilege(const char *user)
    /* Drop from being root down to another user, or exit in disgrace.
     * No helpful error messages, but you can read the exit value.
     * WARNING: This code is run as root. */
{
    struct passwd *pwd;
    uid_t uid;
    gid_t gid;
    
    /* Who do we want to be today? */
    if (user == NULL || user[0] == '\0' || strlen(user) > UT_NAMESIZE) 
	_exit(1); 
    pwd = getpwnam(user);
    if (pwd == NULL) _exit(2);
    uid = pwd->pw_uid;
    gid = pwd->pw_gid;

    /* Even if we are told to, don't run as root */
    if (uid == 0 || gid == 0) _exit(3);

    /* Make the change */
    if (setgroups(0, NULL) != 0) _exit(4);
    if (setgid(gid) != 0) _exit(5);
    if (setuid(uid) != 0) _exit(6);
    
    /* Double-check */
    if (getuid() != uid) _exit(7);
    if (geteuid() != uid) _exit(8);
    if (getgid() != gid) _exit(9);
    if (getegid() != gid) _exit(10);

    /* Safe now */
    return;
}



/*
 *  Non-root functions
 */

static void serviceloop(int socket50)
    /* Central scheduler */
{
    char incoming[UT_NAMESIZE+5]; /* 4 bytes of header; username; null */
    ssize_t r;
    struct sockaddr_in sockaddr;
    socklen_t sa_size;
    struct passwd *pwd;
    struct dirent *dep;
    DIR *new_dp, *cur_dp;
    time_t now, slot = 0;
    int requests_this_slot = 0;

    while(1) {
    get_next_request:

	/* Put a known elephant in Cape Town */
	memset(incoming, 0, UT_NAMESIZE+5); 

	/* Wait for a request */
	sa_size = sizeof(sockaddr);
	memset(&sockaddr, 0, sizeof(sockaddr));
	r = recvfrom(socket50, incoming, UT_NAMESIZE+4, 0,
		     (struct sockaddr *)&sockaddr, &sa_size);
	
	/* Throttle back if we're working too hard */
	if (time(&now) / RATE_INTERVAL != slot) {
	    slot = now / RATE_INTERVAL;
	    requests_this_slot = 1;
	} else if (++requests_this_slot > RATE_COUNT) {
	    sleep (RATE_INTERVAL);
	    continue; /* while(1) */
	}

	/* Parse and respond: header should be all zeroes */
	if (r > 4 
	    && incoming[0] == '\0' && incoming[1] == '\0'
	    && incoming[2] == '\0' && incoming[3] == '\0' 
	    && incoming[4] != '\0')
	{

	    /* Who did they ask for? */
	    cur_dp = new_dp = NULL;
	    if (strlen(incoming+4) != strspn(incoming+4, ACCEPTABLE_CHARS)
		|| (pwd = getpwnam(incoming+4)) == NULL 
		|| pwd->pw_dir == NULL
		|| chdir(pwd->pw_dir) != 0
		|| chdir("./Maildir") != 0
		|| (new_dp = opendir("./new")) == NULL
		|| (cur_dp = opendir("./cur")) == NULL)
	    {
		/* RFC 1339 says we should report "no mail" for any user
		 * who doesn't exist or who hasn't opted in (by letting us
		 * read their maildir). */

		if (new_dp != NULL) closedir(new_dp);
		sendto(socket50, no_mail, 12, 0,
		       (struct sockaddr *)&sockaddr, sa_size);
		continue; /* while(1) */
	    }
		
	    /* The protocol is designed for flat-file mailboxes where
	     * we could just read the mtime and atime.  With maildirs
	     * it's not so easy: we could make an approximation based
	     * in directory mtimes, but there's no way of telling the
	     * last time a maildir was read, if we also want to know
	     * if it's empty.  So, rather than make something that's
	     * almost but not quite the real thing, we use the other
	     * response format, without timestamps.  */

	    while((dep = readdir(new_dp)) != NULL) {
		if (strcmp(dep->d_name, ".") && strcmp(dep->d_name, "..")) {

		    /* New mail */
		    sendto(socket50, new_mail, 12, 0,
			   (struct sockaddr *)&sockaddr, sa_size);
		    closedir(new_dp);
		    closedir(cur_dp);
		    goto get_next_request;
  		}
	    }

	    while((dep = readdir(cur_dp)) != NULL) {
		if (strcmp(dep->d_name, ".") && strcmp(dep->d_name, "..")) {

		    /* Old mail */
		    sendto(socket50, old_mail, 12, 0,
			   (struct sockaddr *)&sockaddr, sa_size);
		    closedir(new_dp);
		    closedir(cur_dp);
		    goto get_next_request;
		}
	    }

	    /* No mail in this maildir */
	    sendto(socket50, no_mail, 12, 0,
		   (struct sockaddr *)&sockaddr, sa_size);
	    closedir(new_dp);
	    closedir(cur_dp);
	}
    }
}


static void failure(const char *why, ...)
    /* Print an explanatory message and die */
{
    va_list ap;
    fprintf (stderr, "remailckd: ");

    va_start (ap, why);
    vfprintf (stderr, why, ap);
    va_end (ap);

    fprintf (stderr, "\n");
    exit(-1);
}


/*
 *  EOF
 */
